package filesystem

import kotlin.io.path.Path

// Note sealed allows for simpler logic in FSCreator by guaranteeing FSFile and FSFolder are the only possible FSEntries
// (as we expect), and it also implicitly makes the class abstract as required.
sealed class FSEntry(val name: String) {
    init {
        val p = Path(name)
        // Only allow single filenames (no paths or relative references (e.g. ".."))
        if (p.toList().size != 1 || p.fileName != p.toFile().canonicalFile.toPath().fileName) {
            throw InvalidEntryNameException(name)
        }
    }
}

class FSFile(name: String, val content: String) : FSEntry(name)

class FSFolder(name: String, val entries: List<FSEntry>) : FSEntry(name) {
    /**
     * Check whether a folder is cyclic.
     */
    fun isCyclic(): Boolean {
        val seen = listOf(this).toHashSet<FSEntry>()
        val queue = ArrayDeque(entries)
        while (queue.isNotEmpty()) {
            val entry = queue.removeFirst()
            if (!seen.add(entry)) {
                return true
            }
            if (entry is FSFolder) {
                queue.addAll(entry.entries)
            }
        }
        return false
    }

    /**
     * Check whether a folder contains multiple entries with the same name.
     */
    fun hasDuplicateNames(): Boolean {
        val seen = HashSet<String>()
        return entries.any { !seen.add(it.name) }
    }

    internal fun deepHasDuplicateNames(): Boolean {
        val queue = ArrayDeque(listOf(this))
        while (queue.isNotEmpty()) {
            val entry = queue.removeFirst()
            if (entry.hasDuplicateNames()) {
                return true
            }
            queue.addAll(entry.entries.mapNotNull { it as? FSFolder })
        }
        return false
    }
}

class InvalidEntryNameException(name: String) : Exception("Invalid FSEntry name: '$name'")