package filesystem

import org.junit.jupiter.api.*
import kotlin.test.assertFalse
import kotlin.test.assertTrue

class FSEntryTest {
    @Test
    fun `valid name entries`() {
        assertDoesNotThrow("should construct FSFile and FSFolder without throwing") {
            FSFile("A file with a name.tar.xz", "Contents")
            FSFolder(".a folder with a name", listOf())
        }
    }

    @Test
    fun `invalid name entries`() {
        assertThrows<InvalidEntryNameException> {
            FSFile("File/here", "Contents")
        }
        assertThrows<InvalidEntryNameException> {
            FSFolder("Folder/here", listOf())
        }
        assertThrows<InvalidEntryNameException> {
            FSFolder(".", listOf())
        }
        assertThrows<InvalidEntryNameException> {
            FSFolder("/", listOf())
        }
    }

    @Test
    fun `non-cyclic folder`() {
        val folder =
            FSFolder(
                "folder",
                listOf(
                    FSFolder("folder", listOf()),
                    FSFile("text.txt", "Hello!"),
                ),
            )
        assertFalse(folder.isCyclic())
    }

    @Test
    fun `cyclic folder`() {
        val files = mutableListOf<FSEntry>()
        val folder = FSFolder("folder", files)
        files.add(folder)
        assertTrue(folder.isCyclic())
    }

    @Test
    fun `long cyclic folder`() {
        val files = mutableListOf<FSEntry>()
        val folder1 = FSFolder("folder", files)
        val folder2 = FSFolder("folder2", listOf(folder1))
        val folder3 = FSFolder("folder3", listOf(folder2))
        val folder4 = FSFolder("folder4", listOf(folder3))
        files.add(folder4)
        assertTrue(folder1.isCyclic())
        assertTrue(folder2.isCyclic())
        assertTrue(folder3.isCyclic())
        assertTrue(folder4.isCyclic())
    }

    @Test
    fun `no duplicate names folder`() {
        val folder =
            FSFolder(
                "folder",
                listOf(
                    FSFile("README.md", "# Test File"),
                    FSFile("hello-world.txt", "Hello World!"),
                    FSFolder(
                        "src",
                        listOf(
                            FSFile("README.md", "# Source files"),
                            FSFile("solution-1.py", "print('1 + 1 = 1')"),
                            FSFile("solution-2.py", "print('1 + 1 = 1')"),
                        ),
                    ),
                    FSFolder("tmp", listOf()),
                ),
            )
        assertFalse(folder.hasDuplicateNames())
        assertFalse(folder.deepHasDuplicateNames())
    }

    @Test
    fun `shallow duplicate names folder`() {
        val folder =
            FSFolder(
                "folder",
                listOf(
                    FSFile("README.md", "# Test File"),
                    FSFile("hello-world.txt", "Hello World!"),
                    FSFolder(
                        "src",
                        listOf(
                            FSFile("README.md", "# Source files"),
                            FSFile("solution-1.py", "print('1 + 1 = 1')"),
                            FSFile("solution-2.py", "print('1 + 1 = 1')"),
                        ),
                    ),
                    FSFolder("tmp", listOf()),
                    FSFile("tmp", "A temporary file"),
                ),
            )
        assertTrue(folder.hasDuplicateNames())
        assertTrue(folder.deepHasDuplicateNames())
    }

    @Test
    fun `deep duplicate names folder`() {
        val folder =
            FSFolder(
                "folder",
                listOf(
                    FSFile("README.md", "# Test File"),
                    FSFile("hello-world.txt", "Hello World!"),
                    FSFolder(
                        "src",
                        listOf(
                            FSFile("README.md", "# Source files"),
                            FSFile("solution-1.py", "print('1 + 1 = 1')"),
                            FSFile("solution-1.py", "print('1 + 1 = 1')"),
                        ),
                    ),
                    FSFolder("tmp", listOf()),
                ),
            )
        assertFalse(folder.hasDuplicateNames())
        assertTrue(folder.deepHasDuplicateNames())
    }
}